# SnapMatcher.py
# Copyright 2005, 2006, 2007 Matt Heinzen

# This file is part of SnapMatcher 0.4.
#
# SnapMatcher is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# SnapMatcher is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SnapMatcher; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import Image
from scipy import *
import os, sys, re, stat, time
from math import *

class ImageRecord:
	"""
	Stores the path, file metadata, and signature vector for a single image.
	"""
	def __init__(self, path, fileSize, fileModTime, recModTime, coefs4x4):
		self.path = path
		self.fileSize = fileSize
		self.fileModTime = fileModTime
		self.recModTime = recModTime
		self.coefs4x4 = coefs4x4


class ImageDB(list):
	"""
	Stores a list of ImageRecords generated for a set of image files. Also
	stores the list of file extensions used to build the database, so that later
	DB updates can automatically identify new images using the same extensions.
	"""
	def __init__(self, rootDir, fileExts, updateTime):
		self.rootDir = rootDir
		self.fileExts = fileExts
		self.updateTime = updateTime


class MatchList(list):
	"""
	A list of image match records generated during full DB match analysis.
	"""
	def __init__(self, rootDir, analysisTime):
		self.rootDir = rootDir
		self.analysisTime = analysisTime


def calcImageSig(img, steps):
	"""Calculates a signature vector for the given image and grid resolution."""
	return list(img.resize((steps, steps), Image.ANTIALIAS).getdata())


def createFileExtRegex(fileExts):
	"""
	Returns a regular expression to determine if a filename matches any of the
	provided file extensions, or any file if no extensions are provided. The
	pattern will apply to lowercase filenames only.
	"""
	if len(fileExts) == 0:
		return re.compile("")
	else:
		pattern = "\.("
		for ext in fileExts:
			pattern = pattern + ext + "|"
		pattern = pattern[0:-1] + ")$"
		return re.compile(pattern)


def createHiddenDirRegex():
	"""Returns a regex to detect hidden (dot-prefixed) directories."""
	if os.sep == '\\': pattern = r"\\\."
	else: pattern = os.sep + "\."
	return re.compile(pattern)


def addImageToDB(imageDB, path, timestamp, statmode):
	"""
	Loads the image file at the given path, generates a sgnature, and adds a
	record to the given image DB.
	"""
	img = Image.open(path)
	if img.getbands() != ('L'):
		img = img.convert('L')
	imageDB.append(ImageRecord(path[len(imageDB.rootDir):],
		statmode[stat.ST_SIZE], statmode[stat.ST_MTIME], timestamp,
		calcImageSig(img, 4)))


def createImageDB(rootDir, dirList, fileExts = []):
	"""
	Returns a dictionary of image file paths to ImageRecords for all image files
	found in the given directories.

	Only files matching one of the provided file extensions will be processed if
	a list of extensions is provided.
	"""

	# create regexes for identifying image files and hidden directories
	extRE = createFileExtRegex(fileExts)
	hiddenRE = createHiddenDirRegex()

	dbCreateTime = int(time.time())
	imageDB = ImageDB(rootDir, fileExts, dbCreateTime)

	for directory in dirList:
		for root, dirs, files in os.walk(directory):
			if re.search(hiddenRE, root) == None:
				print "scanning %s (%d files)" % (root, len(files))

				for name in files:
					if re.search(extRE, name.lower()) != None:
						path = os.path.join(root, name)
						try:
							addImageToDB(imageDB, path, dbCreateTime,
								os.stat(path))
						except (IOError, OSError):
							print "Unable to open image %s" % (path)

	return imageDB


def updateImageDB(imageDB, updateMatches):
	"""
	Returns an image database, efficiently updated from the given existing
	database by only generating signatures for new or modified images, updating
	paths for moved images, and removing records for images which no longer
	exist.

	Optionally returns a dictionary of old->new paths for moved files and a list
	of file which have been modified or removed from the DB for use in updating
	an existing set of image matches.
	"""
	# TODO: update match lists in sync with image DB
	# TODO: improve efficiency by reducing number of dictionary lookups
	#   (possibly using exception mechanism to catch key errors since we expect
	#   most images to still exist, unmodified)

	for record in imageDB: record.confirmed = False

	# map unique file identifiers (including filename but not full directory
	#   path) to image records
	fileMap = dict(zip([(x.path.split(os.sep)[-1], x.fileSize, x.fileModTime)
		for x in imageDB], imageDB))

	# choose the real dictionary update function only when updating matches in
	#   sync with DB, otherwise choose the NIL function (to simplify the loop
	#   code and eiliminate redundant boolean evaluations)
	def updatePathReal(old, new):
		oldToNew[old] = new
	def updatePathNull(old, new):
		pass
	if updateMatches:
		oldToNew = {}
		updatePath = updatePathReal
	else:
		updatePath = updatePathNull

	# confirm existence of all files currently in DB (modified files will be
	#   left as unconfirmed so that the current record will be discarded and a
	#   new one created during the directory scan)
	for rec in imageDB:
		fullPath = imageDB.rootDir + rec.path
		if os.path.exists(fullPath):
			mode = os.stat(fullPath)
			rec.confirmed = rec.fileSize == mode[stat.ST_SIZE] and \
				rec.fileModTime == mode[stat.ST_MTIME]
			if rec.confirmed:
				updatePath(rec.path, rec.path)
			# DEBUG:
			else: print "'%s' modified" % (fullPath)
		# DEBUG:
		else: print "'%s' not found" % (fullPath)

	# create regexes for identifying image files and hidden directories
	extRE = createFileExtRegex(imageDB.fileExts)
	hiddenRE = createHiddenDirRegex()

	dbUpdateTime = int(time.time())

	# scan directories and update DB
	# TODO: add real support for multiple root directories
	for directory in [imageDB.rootDir]:
		for root, dirs, files in os.walk(directory):
			if re.search(hiddenRE, root) == None:
				# handle progress output
				print "scanning %s (%d files)" % (root, len(files))

				for name in files:
					if re.search(extRE, name.lower()) != None:
						path = os.path.join(root, name)
						mode = os.stat(path)
						key = (path.split(os.sep)[-1], mode[stat.ST_SIZE],
							mode[stat.ST_MTIME])

						# previously stored image present and unmodified, but
						#   may have been moved
						if key in fileMap:
							# file matches DB record, but not found in previous
							#   location so file has been moved to a new
							#   directory and just needs its path updated
							if not fileMap[key].confirmed:
								updatePath(fileMap[key].path,
									path[len(imageDB.rootDir):])

								# DEBUG:
								print "%s moved to %s" % (fileMap[key].path,
									path[len(imageDB.rootDir):])

								fileMap[key].path = path[len(imageDB.rootDir):]
								fileMap[key].confirmed = True
						# image not in DB, add a new record
						else:
							try:
								addImageToDB(imageDB, path, dbUpdateTime, mode)
								imageDB[-1].confirmed = True
							except IOError:
								print "Unable to open image %s" % (path)

	# remove old entries which have not been confirmed or updated
	newImageDB = ImageDB(imageDB.rootDir, imageDB.fileExts, dbUpdateTime)
	records = [x for x in imageDB if x.confirmed]
	newImageDB += records

	if updateMatches:
		return newImageDB, oldToNew
	else:
		return newImageDB


def writeImageDBFile(imageDB, outFile):
	"""Write the image DB to a text file."""
	outFile.write(imageDB.rootDir + "\n" + " ".join(imageDB.fileExts) + "\n"
		+ str(imageDB.updateTime) + "\n\n")
	for rec in imageDB:
		outFile.write(rec.path + "*" + str(rec.fileSize) + " "
			+ str(rec.fileModTime) + " " + str(rec.recModTime) + " "
			+ " ".join(map(str, rec.coefs4x4)) + "\n")


def readImageDBFile(inFile):
	"""
	Reads an *.imagedb file into a list of ImageRecords.
	"""
	rootDir = inFile.readline().rstrip()
	fileExts = inFile.readline().split()
	updateTime = int(inFile.readline())
	inFile.readline()
	imageDB = ImageDB(rootDir, fileExts, updateTime)
	for line in inFile.readlines():
		path, rest = line.split('*')
		nums = rest.split()
		imageDB.append(ImageRecord(path, int(nums[0]), int(nums[1]),
			int(nums[2]), map(int, rest.split()[3:])))
	return imageDB


def tryReadImageDBFile(imageDBFilename):
	"""Returns an image DB read from the given filename, exiting on failure."""
	try:
		imageDBFile = open(imageDBFilename, "r")
		imageDB = readImageDBFile(imageDBFile)
		imageDBFile.close()
	except IOError:
		print "ERROR: unable to read image database file '%s'" \
			% (imageDBFilename)
		sys.exit()
	return imageDB


def choose2(a):
	"""Returns a generator for enumerating n-choose-2 unordered list pairs."""
	return ((x, y) for i, x in enumerate(a) for j, y in enumerate(a) if i < j)


def addMatchesToList(matchList, imageDB, minCorTol = 0.95, verbose = 1):
	"""
	Generates a list of similar images from candidates in the image list based
	on 4x4 grid signature correlation, and adds to the given match list.

	To build a new match list, provide an empty list with analysisTime set to
	zero.
	"""

	oldSize = len(matchList)
	totalComps = 0

	analysisTime = int(time.time())

	if verbose == 1:
		print "starting scan of %d images..." % (len(imageDB))

	# compare new images to old
	for img1 in (x for x in imageDB if x.recModTime > matchList.analysisTime):
		for img2 in (x for x in imageDB if x.recModTime <= matchList.analysisTime):
			cor = abs(stats.pearsonr(img1.coefs4x4, img2.coefs4x4)[0])
			if cor > minCorTol:
				matchList.append((cor, img1.path, img2.path))
			totalComps += 1

	# compare new images to each other
	newImages = [x for x in imageDB if x.recModTime > matchList.analysisTime]
	for img1, img2 in choose2(newImages):
		cor = abs(stats.pearsonr(img1.coefs4x4, img2.coefs4x4)[0])
		if cor > minCorTol:
			matchList.append((cor, img1.path, img2.path))
		totalComps += 1

	matchList.analysisTime = analysisTime

	if verbose == 1:
		print "Added %d new matches, %d total, using %d comparisons" \
			% (len(matchList) - oldSize, len(matchList), totalComps)


def updateMatchList(matchList, oldToNew):
	"""
	Returns a new a match list, updating old paths for moved files and removing
	matches with files that have been modified or removed from the database
	image set.
	"""
	newMatchList = MatchList(matchList.rootDir, matchList.analysisTime)
	for match in matchList:
		try:
			newMatchList.append((match[0], oldToNew[match[1]],
				oldToNew[match[2]]))
		except KeyError:
			pass
	return newMatchList


def writeMatchListFile(matchList, outFile):
	"""Writes a list a matched image pairs to a text file."""
	outFile.write(matchList.rootDir + "\n" + str(matchList.analysisTime)
		+ "\n\n")
	for elem in matchList:
		outFile.write(str(elem[0]) + "\n" + elem[1] + "\n" + elem[2] + "\n\n")


def readMatchListFile(inFile):
	"""Reads match pairs from the given match list file into a MatchList."""
	def readMatchLines(inFile):
		cor = inFile.readline().rstrip()
		if len(cor) == 0:
			return None
		else:
			cor = float(cor)
			name1 = inFile.readline().rstrip()
			name2 = inFile.readline().rstrip()
			empty = inFile.readline().rstrip()
			return (cor, name1, name2)

	rootDir = inFile.readline().rstrip()
	analysisTime = int(inFile.readline())
	inFile.readline()

	matchList = MatchList(rootDir, analysisTime)

	match = readMatchLines(inFile)
	while match != None:
		matchList.append(match)
		match = readMatchLines(inFile)

	return matchList


def tryReadMatchListFile(matchFilename):
	try:
		matchFile = open(matchFilename, "r")
		matchList = readMatchListFile(matchFile)
		matchFile.close()
	except IOError:
		print "ERROR: unable to read matches file '%s'" % (matchFilename)
		sys.exit()
	return matchList


def buildDB(groupName, rootDir, fileExts):
	"""
	Builds a new image DB with the given name from files under the given
	directory and matching the given file extensions, and writes the DB to
	'groupName.imagedb' in the current working directory.
	"""

	imageDBFilename = groupName + ".imagedb"

	print "Building image database..."
	print "DIR=" + rootDir
	print "EXTS=" + str(fileExts)

	try:
		imageDBFile = open(imageDBFilename, "w")
		imageDB = createImageDB(rootDir, [rootDir], fileExts)
		writeImageDBFile(imageDB, imageDBFile)
		imageDBFile.close()
	except IOError:
		print "ERROR: unable to write image database file '%s'" \
			% (imageDBFilename)
		sys.exit()

	print str(len(imageDB)) + " images analyzed"


def updateDB(groupName):
	"""
	Updates an existing image DB file for new, modified, or removed images in
	the DB directories. Also updates an existing matches file by updating
	matches with mode files and removing matches for deleted or modified files.
	"""

	imageDBFilename = groupName + ".imagedb"
	imageDB = tryReadImageDBFile(imageDBFilename)

	matchFilename = groupName + ".matches"

	# update matches file in addition to updating the DB itself
	if os.path.exists(matchFilename):
		matchList = tryReadMatchListFile(matchFilename)

		newImageDB, oldToNew = updateImageDB(imageDB, True)
		newMatchList = updateMatchList(matchList, oldToNew)
		newMatchList.sort(reverse = True)

		# write new matches file, deleting old only after new file is created
		try:
			matchFile = open(matchFilename + "_temp", "w")
			writeMatchListFile(newMatchList, matchFile)
			matchFile.close()
			os.remove(matchFilename)
			os.rename(matchFilename + "_temp", matchFilename)
		except (IOError, OSError):
			print "ERROR: unable to write matches file '%s'" % (matchFilename)
			sys.exit()

	# no matches file, update image DB only
	else:
		newImageDB = updateImageDB(imageDB, False)

	# write new DB, deleting old only after new file is created
	try:
		imageDBFile = open(imageDBFilename + "_temp", "w")
		writeImageDBFile(newImageDB, imageDBFile)
		imageDBFile.close()
		os.remove(imageDBFilename)
		os.rename(imageDBFilename + "_temp", imageDBFilename)
	except (IOError, OSError):
		print "ERROR: unable to write image database file '%s'" \
			% (imageDBFilename)
		sys.exit()


def findDups(groupName, minCorTol):
	"""
	Loads the image DB with the given group name and identifies all image pairs
	with signature correlation greater than the given tolerance. The list of
	resulting pairs is written to 'groupName.matches' in the current working
	directory.
	"""
	# TODO: improve efficiency by first only reading file headers to compare
	#   timestamps

	imageDBFilename = groupName + ".imagedb"
	imageDB = tryReadImageDBFile(imageDBFilename)

	matchFilename = groupName + ".matches"

	# read the existing matches file and update it with new matches
	if os.path.exists(matchFilename):
		matchList = tryReadMatchListFile(matchFilename)

		if matchList.analysisTime >= imageDB.updateTime:
			print "matches file '%s' is already up to date" % (matchFilename)
		else:
			# write new matches file, deleting old only after new file is
			#   created
			try:
				matchFile = open(matchFilename + "_temp", "w")
				addMatchesToList(matchList, imageDB, minCorTol)
				matchList.sort(reverse = True)
				writeMatchListFile(matchList, matchFile)
				matchFile.close()
				os.remove(matchFilename)
				os.rename(matchFilename + "_temp", matchFilename)
			except (IOError, OSError):
				print "ERROR: unable to write matches file '%s'" \
					% (matchFilename)
				sys.exit()
	# test all image pairs and output to a new matches file
	else:
		try:
			matchFile = open(matchFilename, "w")
			matchList = MatchList(imageDB.rootDir, 0)
			addMatchesToList(matchList, imageDB, minCorTol)
			matchList.sort(reverse = True)
			writeMatchListFile(matchList, matchFile)
			matchFile.close()
		except IOError:
			print "ERROR: unable to write matches file '%s'" % (matchFilename)
			sys.exit()


def matchImages(groupName, imagePaths, minCorTol):
	"""
	Find matches for the given list of image files against the image DB with the
	given group name.
	"""

	# create a temporary image DB for the given files, aborting if any images
	#   fail to be read (so that users can correct possible misspellings before
	#   a full DB load and scan is performed)
	tempDB = ImageDB("", "", 0)
	for path in imagePaths:
		try:
			addImageToDB(tempDB, path, 0, os.stat(path))
		except (IOError, OSError):
			print "ERROR: unable to open image '%s', exiting" % (path)
			sys.exit()

	imageDBFilename = groupName + ".imagedb"
	imageDB = tryReadImageDBFile(imageDBFilename)

	for img1 in tempDB:
		for img2 in imageDB:
			cor = abs(stats.pearsonr(img1.coefs4x4, img2.coefs4x4)[0])
			if cor > minCorTol:
				print "match with correlation %f:" % (cor)
				print "  new:   %s" % (img1.path)
				print "  in DB: %s" % (img2.path)


def printUsageAndExit():
	"""Display valid command line arguments for SnapMatcher."""
	print "USAGES:"
	print "    SnapMatcher.py --builddb <GROUPNAME> <ROOTDIR> [EXT 1] " \
		"[EXT 2] ..."
	print "    SnapMatcher.py --updatedb <GROUPNAME>"
	print "    SnapMatcher.py --finddups <GROUPNAME> [TOLERANCE]"
	print "    SnapMatcher.py --match <GROUPNAME> <TOLERANCE> <FILE 1> " \
		"[FILE 2] [FILE 3] ..."
	print "if no extensions specified for --builddb the following " \
		"default extensions will be used:", defExts
	print "if no minimum correlation tolerance specified for --finddups then " \
		"%.2f will be used" % (defMinCorTol)
	sys.exit()


def fileExtsFromString(fileExtStr = ""):
	"""
	Returns a list of file extensions parsed from the user supplied string.

	Extensions are separated, converted to lowercase, and stripped of all
	separator and wildcard characters.
	"""
	exts = re.split("\W+", fileExtStr.lower())
	if len(exts) == 1 and len(exts[0]) == 0:
		exts = []
	return exts


if __name__ == "__main__":
	defExts = "jpg jpeg jpe png gif tif tiff bmp pcx ppm xpm xbm"
	defMinCorTol = 0.95

	numArgs = len(sys.argv)

	# not enough parameters
	if numArgs < 2: printUsageAndExit()

	# build a new image DB
	elif sys.argv[1] == "--builddb":
		if numArgs < 4: printUsageAndExit()

		# convert root search directory to absolute path
		groupName = sys.argv[2]
		rootDir = os.path.abspath(sys.argv[3]) + os.path.sep

		# extract the valid file extensions to check for
		if numArgs > 4:
			fileExts = fileExtsFromString(" ".join(sys.argv[4:]))
		else: fileExts = defExts.split()

		buildDB(groupName, rootDir, fileExts)

	# update an existing image DB
	elif sys.argv[1] == "--updatedb":
		if numArgs != 3: printUsageAndExit()

		groupName = sys.argv[2]

		updateDB(groupName)

	# find all duplicate image pairs in an image DB
	elif sys.argv[1] == "--finddups":
		if numArgs < 3 or numArgs > 4: printUsageAndExit()

		groupName = sys.argv[2]
		if numArgs > 3:
			try:
				minCorTol = float(sys.argv[3])
			except ValueError:
				minCorTol = -1.0
		else: minCorTol = defMinCorTol

		if minCorTol <= 0.0 or minCorTol >= 1.0:
			print "invalid minimum correlation tolerance '%s' specified" \
				% (sys.argv[3])
			print "value must be a number between 0.0 and 1.0"
			sys.exit()

		findDups(groupName, minCorTol)

	# find matches between a short list of image files and images in an existing
	#   image DB
	elif sys.argv[1] == "--match":
		if numArgs < 4: printUsageAndExit()

		groupName = sys.argv[2]
		try:
			minCorTol = float(sys.argv[3])
		except ValueError:
			minCorTol = -1.0

		if minCorTol <= 0.0 or minCorTol >= 1.0:
			print "invalid minimum correlation tolerance '%s' specified" \
				% (sys.argv[3])
			print "value must be a number between 0.0 and 1.0"
			sys.exit()

		matchImages(groupName, sys.argv[4:], minCorTol)

	# invalid command
	else: printUsageAndExit()

